/* ------------------------------------------------------------------------
 * Program : Lockexec - lockfile management and control.
 *
 * Version  : 0.3.1
 * Author   : Paul Matthews
 * Date     : 2000 April 23
 * Comments : This one actaully appears to work ...
 *
 * ------------------------------------------------------------------------
 * Copyright (C) Paul Matthews 1997,1998,1999,2000
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ------------------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>


void usage( char *name )
{
  fprintf( stderr, "Usage: %s [-share] [-nowait]", name );
  fprintf( stderr, "file.lock cmd [arg]...\n" );
  fprintf( stderr, "Where: --share     shared lock\n" );
  fprintf( stderr, "       --nowait    don't wait for the lock\n" ); 
  fprintf( stderr, "       file.lock   lock file\n" );
  fprintf( stderr, "       cmd         command to executed\n" );
  fprintf( stderr, "       arg...      arguments to the command\n" );
  exit(-1);
}

int main( int argc, char *argv[] )
{
  char        *progname;
  char        *lockname;
  int          fdlock;
  FILE*        filelock;
  int          argn;
  int          share = 0;
  int          nowait= 0;
  struct flock sflock;
  int          pid;
  int          status;

  /* who the hell are we */ 
  progname = argv[0];

  /* called without arguments, print the help*/
  if( argc == 1 ) 
    usage(progname);
  
  /* look for switches */
  for( argn=1; argn<argc; argn++ )
  { 
    if( strcmp("--share",argv[argn]) == 0 )
      share = 1;
    else if ( strcmp("--nowait",argv[argn]) == 0 )
      nowait = 1;
    else
      break;
  }
  
  /* there must be at least two more arguments, the name 
   * of the lockfile and the command to run. */
  if( (argc-argn) < 2 )
    usage(progname);

  /* get the lockfile name */
  lockname = argv[argn++];
  
  /* open the lockfile, create if it does not exist. */
  fdlock =   open( lockname, O_RDWR|O_CREAT, 0660 );
  if( fdlock == -1 )
  {
    fprintf( stderr, "%s: Could not open lockfile '%s' because %s.\n", 
             progname, lockname, strerror(errno) );
    usage(progname);
  }
  
  /* obtain a lock */
  sflock.l_type   = share?F_RDLCK:F_WRLCK;
  sflock.l_whence = 0;
  sflock.l_start  = 0;
  sflock.l_len    = 0;
  while(1)
  {
    if( fcntl( fdlock, nowait?F_SETLK:F_SETLKW, &sflock) == -1 )
    {
      if( errno==EACCES || errno==EAGAIN )
        exit(-1);
      if( errno==EINTR )
        continue;
      fprintf( stderr, 
               "%s: Could not lock lockfile '%s' because "
	       "unexpectedly '%s'.\n",
               progname, lockname, strerror(errno) );
      usage(progname);
    }    
    break;
  }
  
  /* fork a child to run the command, parent will wait. */
  pid = fork();
  if( pid == -1 )
  {
    fprintf( stderr, 
             "%s: Could not fork to execute command because %s.\n", 
             progname, strerror(errno) );
    exit(-1); 
  }
  if( pid == 0 )
  {
    execvp( argv[argn], argv+argn );	/* this should never return */
    fprintf( stderr, 
             "%s: Could not execute command because %s.\n", 
	     progname, strerror(errno) );
    _exit(-1);
  }

  /* write pid in lockfile if exclusive, space otherwise */
  filelock = fdopen( fdlock, "w" );
  if( share )
    fprintf( filelock, "     " );  
  else
    fprintf( filelock, "%5d", (int)pid );
  fflush( filelock );
    
  /* wait for our child process */
  while(1)
  {
    if( waitpid(pid,&status,0) != pid )
    {
      if( errno==EINTR )
        continue;
      fprintf( stderr, 
               "%s: Could not wait for executed program to complete "
	       "because %s.\n",
	       progname, strerror(errno) );
      exit(-1);
    }
    else
      break;  
  }

  /* Put spaces back in the lockfile */
  if( ! share )
  {
    fseek( filelock, 0, SEEK_SET );
    fprintf( filelock, "     " );
    fflush( filelock );
  }
    
  /* Unlock the lock file */
  sflock.l_type = F_UNLCK;
  fcntl( fdlock, F_SETLKW, &sflock );
  
  /* Close the lockfile */
  fclose( filelock );
  
  /* Finish */
  exit(status); 
}
